/*
 * Device Motion Gesture for JavaScript.
 * http://sourceforge.net/projects/jsmotiongesture/
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU General Public License along
 * with the files in this library. If not, see  http://www.gnu.org/licenses
 * Distributed under the GNU Library or Lesser General Public License (LGPL)
 */


if (typeof Motion === "undefined")
	Motion = {};

if (typeof Motion.Gesture === "undefined")
	Motion.Gesture = {};




/**
 * The Model namespace holds utility functions for operations on the POJSOs described above.
 * This approach (using simple beans combined with operation classes) is preferred to using classes
 * with methods because it keeps the data definition simple and the goal is to have gestures defined
 * as simple JSON constructs which can be quickly pasted into applications.
 */
Motion.Gesture.Model = {};




/*
 * In the documentation you will see references to POJSOs with the following names.
 * Their structure is defined below.
 *
 * Gesture
 *        .id                     {string}     A unique name for this gesture.
 *        .style.lineColour       {string}     The web colour name/code for the line drawn to depict the motion of this gesture.
 *        .style.lineOpacity      {number}     The opacity for the line in the range 0 to 1.
 *        .style.paleColour       {string}     The web colour name/code for the pale annotations (control windows) drawn for this gesture.
 *        .signature              {Signature}  Definition of the gesture signature.
 *        .ui.tabColour           {string}     The web colour name/code for the UI element which contains the buttons/controls for this gesture.
 *        .ui.tabElementId        {string}     The id of the DOM element which labels the selected tab for this gesture.
 *
 * Signature
 *          .roots                     [{Control}]  The root controls for the gesture's signature.
 *          .metadata.record.endTime   {number}     The timestamp of when the gesture was recorded.
 *
 * Control
 *        .signature              {Signature}  the parent, signature object
 *        .window.time.start      {number}  timestamp of the control's time window (only defined for a root control)
 *        .window.time.delay      {number}  distance from the previous control, in msecs (not defined for a root control)
 *        .window.time.width      {number}  width of the control window, in msecs
 *        .points                 [{ControlPoint}]  the control points that define the acceptance criteria for this control
 *        ------------------
 *        <<intermediate properties, used during the construction of the signature>>
 *        .timestamp.min          {number}  the minimum timestamp of all points in this control
 *        .timestamp.max          {number}  the maximum timestamp of all points in this control
 *        .next                   {Control} the control that follows on from this one, or null if this is the last control
 *        .previous               {Control} the control that preceded this one, or null if this is a root control
 *
 * ControlPoint
 *             .axis              {string}  name of the axis this control point exists within, "x" "y" "z", "xg", "yg", "zg"
 *             .direction         {number}  the direction of the, see the constants defined by the Motion.Gesture.Record.Control.Generator class
 *             .timestamp         {number}  the timestamp for when this point occurred
 *             .value             {number}  the acceleration value at this point
 *             .window.value.min  {number}  the minimum acceleration value considered to match this point
 *             .window.value.max  {number}  the maximum acceleration value considered to match this point
 */


Motion.Gesture.Model.Control = {};




Motion.Gesture.Model.Control.getStartTime = function( control )
{
	if (control.previous)
	{
		var previousControlMidPoint = Motion.Gesture.Model.Control.getStartTime( control.previous ) + (control.previous.window.time.width / 2);

		// The delay between controls is measured from the centre of the previous control (when the previous control is most expected to be met)
		// to the start of the time window of the next control (when monitoring of the next control should be started).
		return previousControlMidPoint + control.window.time.delay;
	}
	else
	{
		// This is the root control.
		return control.window.time.start;
	}
}


Motion.Gesture.Model.Gesture = {};




/**
 * @param {Motion.Gesture.Model.Gesture} gesture  The gesture object.
 * @returns {boolean}  True if the gesture is defined.
 *                     False if the gesture object is a placeholder entity, waiting
 *                     for a signature to be added to it.
 */
Motion.Gesture.Model.Gesture.isDefined = function( gesture )
{
	return (gesture.signature);
}

Motion.Gesture.Model.Serialisation = {};




/**
 * @param {string} jsonString  The JSON text represents a gesture signature.
 * @returns {Signature}  The gesture signature.
 */
Motion.Gesture.Model.Serialisation.signatureFromJson = function( jsonString )
{
	var signature = JSON.parse( jsonString );

	Motion.Gesture.Model.Serialisation.restoreDeserialisedSignature( signature );

	return signature;
}

/**
 * @param {Signature} signature  The gesture signature.
 * @param {string} indentation  The padding string used to indent the JSON,
 *                              or leave undefined to serialise the JSON on a single line of text.
 * @returns {string}  The JSON representation of the signature.
 */
Motion.Gesture.Model.Serialisation.signatureToJson = function( signature, indentation )
{
	return JSON.stringify( signature, Motion.Gesture.Model.Serialisation.jsonSerialisationFilter, indentation );
}

Motion.Gesture.Model.Serialisation.jsonSerialisationFilter = function( key, value )
{
	// Ignore the previous property of the Control objects, which would otherwise cause a cyclic reference.
	// When deserialising the JSON the 'previous' property can be restored using the restoreDeserialisedSignature function.
	if (key === "previous")
	{
		return undefined;
	}

	// Ignore the signature property of the Control objects, which would otherwise cause a cyclic reference.
	// When deserialising the JSON the 'signature' property can be restored using the restoreDeserialisedSignature function.
	if (key === "signature")
	{
		return undefined;
	}

	return value;
}

/**
 * Post-process a signature object obtained from a simple call to JSON.parse.
 * This will restore properties that were not able to be written to the JSON.
 *
 * @param {Signature}  The signature object.
 */
Motion.Gesture.Model.Serialisation.restoreDeserialisedSignature = function( signature )
{
	for (var i=0; i<signature.roots.length; i++)
	{
		var control = signature.roots[i];

		while (control.next)
		{
			// Restore the 'previous' property.
			control.next.previous = control;

			// Restore the 'signature' property.
			control.signature = signature;

			control = control.next;
		}
	}
}
