/*
 * Device Motion Gesture for JavaScript.
 * http://sourceforge.net/projects/jsmotiongesture/
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU General Public License along
 * with the files in this library. If not, see  http://www.gnu.org/licenses
 * Distributed under the GNU Library or Lesser General Public License (LGPL)
 */


if (typeof Motion === "undefined")
	Motion = {};

if (typeof Motion.Gesture === "undefined")
	Motion.Gesture = {};




/**
 * The recognition package contains the classes which are used to detect
 * gestures from motion data - either live or recorded.
 */
Motion.Gesture.Recognition = {};


/**
 * Basic strategy class for checking the conditions of a control point, control and array of controls.
 */
Motion.Gesture.Recognition.ConditionComparator = function()
{
}




Motion.Gesture.Recognition.ConditionComparator.prototype.isRootControlConiditionMet = function( rootControls, data )
{
	for (var i=0; i<rootControls.length; i++)
	{
		if (this.isControlConiditionMet( rootControls[i], data ) === false)
		{
			return false;
		}
	}

	// All root control conditions were met.
	return true;
}

Motion.Gesture.Recognition.ConditionComparator.prototype.isControlConiditionMet = function( control, data )
{
	for (var i=0; i<control.points.length; i++)
	{
		if (this.isControlPointConiditionMet( control.points[i], data ) === false)
		{
			return false;
		}
	}

	// All control point conditions were met.
	return true;
}

Motion.Gesture.Recognition.ConditionComparator.prototype.isControlPointConiditionMet = function( controlPoint, data )
{
	var value = data[controlPoint.axis];
	if (value)
	{
		return (value > controlPoint.window.value.min  &&  value < controlPoint.window.value.max);
	}
	else
	{
		// There is no value for the axis monitored by this control point. It can't be a match.
		return false;
	}
}


/**
 * The ControlThread processes motion data and checks whether it matches a given control.
 * After creating the thread, specifying the control to be mointored, you must start the thread
 * giving the start time.
 * Keep calling the process function to process the motion data. After each call to process you
 * should check the status to detect when the thread has expired and whether the control conditions are met.
 *
 * @param {Control} control  The control being monitored by this thread.
 * @param {ConditionComparator} conditionComparator  Strategy for checking whether data matches this control point's conditions.
 */
Motion.Gesture.Recognition.ControlThread = function( control, conditionComparator )
{
	this.control = control;
	this.conditionComparator = conditionComparator;
	this.status = {};
	this.status.expired = false;
	this.status.conditionMet = false;
}




/**
 * @param {number} timestamp  Optional value to set a start time in the past. If omitted the current time is used.
 */
Motion.Gesture.Recognition.ControlThread.prototype.start = function( timestamp )
{
	var startTime = (timestamp) ? timestamp : new Date().getTime();

	this.timeFrame = {};
	this.timeFrame.start = startTime + this.control.window.time.delay;
	this.timeFrame.end = this.timeFrame.start + this.control.window.time.width;
}

/**
 * @param {MotionData} data  The recorded motion data.
 * @returns {boolean}  Whether the status of this thread has changed because of this new data.
 */
Motion.Gesture.Recognition.ControlThread.prototype.process = function( data )
{
	if (data.timestamp < this.timeFrame.start)
	{
		// The control is not yet active.
		return false;
	}

	if (data.timestamp > this.timeFrame.end)
	{
		// The control has expired.
		this.stop();
		return true;
	}

	if (this.conditionComparator.isControlConiditionMet( this.control, data ) === true)
	{
		// The control condition has been met.
		this.status.conditionMet = true;
		this.stop();
		return true;
	}

	// No need to change the status of this thread.
	return false;
}

Motion.Gesture.Recognition.ControlThread.prototype.stop = function()
{
	this.status.expired = true;
}


/**
 * The GestureThread processes motion data and checks whether it matches a given gesture.
 * After creating the thread, specifying the gesture to be mointored, you must start the thread
 * giving the start time.
 * Keep calling the process function to process the motion data. After each call to process you
 * should check the status to detect when the thread has expired and whether the gesture conditions are met.
 *
 * @param {Gesture} gesture  The gesture being monitored by this thread.
 * @param {ConditionComparator} conditionComparator  Strategy for checking whether data matches control point conditions.
 */
Motion.Gesture.Recognition.GestureThread = function( gesture, conditionComparator )
{
	this.gesture = gesture;
	this.conditionComparator = conditionComparator;
	this.controlThreads = [];
	this.status = {};
	this.status.expired = false;
	this.status.conditionMet = false;
}




/**
 * @param {number} timestamp  Optional value to set a start time in the past. If omitted the current time is used.
 */
Motion.Gesture.Recognition.GestureThread.prototype.start = function( timestamp )
{
	var _timestamp = (timestamp) ? timestamp : null;

	for (var i=0; i<this.gesture.signature.roots.length; i++)
	{
		this.startMonitoringControl( this.gesture.signature.roots[i].next, _timestamp );
	}
}

Motion.Gesture.Recognition.GestureThread.prototype.startMonitoringControl = function( control, timestamp )
{
	var _timestamp = (timestamp) ? timestamp : null;

	var thread = new Motion.Gesture.Recognition.ControlThread( control, this.conditionComparator );
	thread.start( _timestamp );

	this.controlThreads[this.controlThreads.length] = thread;
}

/**
 * @param {MotionData} data  The recorded motion data.
 * @returns {boolean}  Whether the status of this thread has changed because of this new data.
 */
Motion.Gesture.Recognition.GestureThread.prototype.process = function( data )
{
	var anyThreadStatusChanged = false;
	for (var i=0; i<this.controlThreads.length; i++)
	{
		var controlThread = this.controlThreads[i];

		if (controlThread.status.expired === false)
		{
			var threadStatusChanged = controlThread.process( data );

			if (threadStatusChanged === true)
			{
				anyThreadStatusChanged = true;

				if (controlThread.status.expired === true  &&  controlThread.status.conditionMet  &&  controlThread.control.next)
				{
					this.startMonitoringControl( controlThread.control.next, data.timestamp );
				}
			}
		}
	}

	if (anyThreadStatusChanged === true)
	{
		return this.updateStatus();
	}
	else
	{
		// None of the sub-threads have changed. No change in this gesture thread.
		return false;
	}
}

Motion.Gesture.Recognition.GestureThread.prototype.updateStatus = function()
{
	var statusChanged = false;

	if (this.allThreadsExpired() === true)
	{
		if (this.allControlsMet() === true)
		{
			this.status.conditionMet = true;
		}

		this.stop();

		statusChanged = true;
	}

	return statusChanged;
}

/**
 * Check if all the sub-threads (control point threads) have expired.
 *
 * @returns {boolean}  True if all the sub-threads have expired.
 */
Motion.Gesture.Recognition.GestureThread.prototype.allThreadsExpired = function()
{
	for (var i=0; i<this.controlThreads.length; i++)
	{
		if (this.controlThreads[i].status.expired === false)
		{
			return false;
		}
	}

	return true;
}

/**
 * Check if all the sub-threads (control threads) found matching motion data.
 *
 * @returns {boolean}  True if all the controls were met.
 */
Motion.Gesture.Recognition.GestureThread.prototype.allControlsMet = function()
{
	for (var i=0; i<this.controlThreads.length; i++)
	{
		if (this.controlThreads[i].status.conditionMet === false)
		{
			return false;
		}
	}

	return true;
}

Motion.Gesture.Recognition.GestureThread.prototype.stop = function()
{
	this.status.expired = true;
}


/**
 * The Monitor class is the main class for recognition.
 * You register gestures with the monitor and pass motion data into the process function.
 * The monitor then calls the trigger functions when a gesture is recognised.
 *
 * @param {ConditionComparator} conditionComparator  Optional strategy for checking whether data matches control point conditions.
 *                                                   If undefined the basic class (Motion.Gesture.Recognition.ConditionComparator) is used.
 */
Motion.Gesture.Recognition.Monitor = function( conditionComparator )
{
	this.conditionComparator = (conditionComparator) ? conditionComparator : new Motion.Gesture.Recognition.ConditionComparator();
	this.gestures = [];
	this.gestureTriggers = {};
	this.gestureThreads = [];
}




Motion.Gesture.Recognition.Monitor.prototype.addGesture = function( gesture, trigger )
{
	// Remove any previous registration for this gesture.
	this.removeGesture( gesture );

	this.gestures[this.gestures.length] = gesture;
	this.gestureTriggers[gesture.id] = trigger;
}

Motion.Gesture.Recognition.Monitor.prototype.removeGesture = function( gesture )
{
	this.stopThreadsForGesture( gesture );

	this.gestureTriggers[gesture.id] = null;

	var gestureIndex = this.gestures.indexOf( gesture );
	if (gestureIndex > -1)
	{
		this.gestures.splice( gestureIndex, 1 );
	}
}

Motion.Gesture.Recognition.Monitor.prototype.stopAllThreads = function()
{
	for (var i=0; i<this.gestures.length; i++)
	{
		this.stopThreadsForGesture( this.gestures[i] );
	}
}

Motion.Gesture.Recognition.Monitor.prototype.stopThreadsForGesture = function( gesture )
{
	for (var i=0; i<this.gestureThreads.length; i++)
	{
		var thread = this.gestureThreads[i];
		if (thread.gesture.id == gesture.id)
		{
			thread.stop();
		}
	}

	this.dereferenceExpiredThreads();
}

Motion.Gesture.Recognition.Monitor.prototype.dereferenceExpiredThreads = function()
{
	var i = 0;
	while (i < this.gestureThreads.length)
	{
		if (this.gestureThreads[i].status.expired === true)
		{
			if (this.gestureThreads[i].status.conditionMet === true)
			{
				var gestureId = this.gestureThreads[i].gesture.id;
				this.gestureTriggers[gestureId]();
			}

			this.gestureThreads.splice( i, 1 );
		}
		else
		{
			i ++;
		}
	}
}

Motion.Gesture.Recognition.Monitor.prototype.process = function( data )
{
	this.processCurrentThreads( data );
	this.processNewThreads( data );
}

Motion.Gesture.Recognition.Monitor.prototype.processCurrentThreads = function( data )
{
	var anyThreadStatusChanged = false;

	for (var i=0; i<this.gestureThreads.length; i++)
	{
		var threadStatusChanged = this.gestureThreads[i].process( data );
		if (threadStatusChanged === true)
		{
			anyThreadStatusChanged = true;
		}
	}

	if (anyThreadStatusChanged === true)
	{
		this.dereferenceExpiredThreads();
	}
}

Motion.Gesture.Recognition.Monitor.prototype.processNewThreads = function( data )
{
	for (var i=0; i<this.gestures.length; i++)
	{
		var gesture = this.gestures[i];
		if (this.conditionComparator.isRootControlConiditionMet( gesture.signature, data ) === true)
		{
			this.startMonitoringGesture( data.timestamp, gesture );
		}
	}
}

Motion.Gesture.Recognition.Monitor.prototype.startMonitoringGesture = function( timestamp, gesture )
{
	var thread = new Motion.Gesture.Recognition.GestureThread( gesture, this.conditionComparator );
	thread.start( timestamp );

	this.gestureThreads[this.gestureThreads.length] = thread;
}
