/*
 * Device Motion Gesture for JavaScript.
 * http://sourceforge.net/projects/jsmotiongesture/
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU General Public License along
 * with the files in this library. If not, see  http://www.gnu.org/licenses
 * Distributed under the GNU Library or Lesser General Public License (LGPL)
 */


if (typeof Motion === "undefined")
	Motion = {};

if (typeof Motion.Gesture === "undefined")
	Motion.Gesture = {};




Motion.Gesture.Record = {};

if (typeof Motion === "undefined")
	Motion = {};

if (typeof Motion.Gesture === "undefined")
	Motion.Gesture = {};

if (typeof Motion.Gesture.Record === "undefined")
	Motion.Gesture.Record = {};




Motion.Gesture.Record.Controls = {};


/**
 *
 */
Motion.Gesture.Record.Controls.Builder = function( controlPointGenerator, controlPointLinker, controlPointMerger, controlSizer, controlCleaner )
{
	this.controlPointGenerator = controlPointGenerator;
	this.controlPointLinker = controlPointLinker;
	this.controlPointMerger = controlPointMerger;
	this.controlSizer = controlSizer;
	this.controlCleaner = controlCleaner;
}




Motion.Gesture.Record.Controls.Builder.prototype.build = function( motionData )
{
	var controlPoints = this.controlPointGenerator.parse( motionData );
	var rootControlPoints = this.controlPointLinker.link( controlPoints );
	var rootControls = this.controlPointMerger.merge( rootControlPoints );
	this.controlSizer.size( rootControls );
	this.controlCleaner.clean( rootControls );

	return { "roots": rootControls,
		 "metadata": motionData.metadata };
}


/**
 * The Cleaner removes properties that have been assigned and used during the building of the gesture
 * but are not needed and should be cleaned-up at the end of the building process.
 */
Motion.Gesture.Record.Controls.Cleaner = function()
{
}




Motion.Gesture.Record.Controls.Cleaner.prototype.clean = function( rootControls )
{
	for (var i=0; i<rootControls.length; i++)
	{
		var control = rootControls[i];

		while (control)
		{
			this.cleanControl( control );

			for (var j=0; j<control.points.length; j++)
			{
				this.cleanControlPoint( control.points[j] );
			}

			control = control.next;
		}
	}
}

Motion.Gesture.Record.Controls.Cleaner.prototype.cleanControl = function( control )
{
	this.removeTimestamp( control );
}

Motion.Gesture.Record.Controls.Cleaner.prototype.cleanControlPoint = function( controlPoint )
{
	this.removeControlPointLinks( controlPoint );
}

Motion.Gesture.Record.Controls.Cleaner.prototype.removeTimestamp = function( control )
{
	control.timestamp = undefined;
}

Motion.Gesture.Record.Controls.Cleaner.prototype.removeControlPointLinks = function( controlPoint )
{
	controlPoint.previous = undefined;
	controlPoint.next = undefined;
}


/**
 * The linker turns the array of control points into a linked list,
 * adding 'previous' and 'next' properties to each control point.
 */
Motion.Gesture.Record.Controls.Linker = function()
{
}




/**
 * Link the control points for the x-y-z axes together
 * and the control points for the xg-yg-zg axes together.
 * The 2 roots for each group of axes are returned in an array.
 *
 * @param controlPoints [object]  an object model containing the control points for the 6 motion axes
 * @return [Array(ControlPoint)]  the root control points
 */
Motion.Gesture.Record.Controls.Linker.prototype.link = function( controlPoints )
{
	var withoutGravityPoints = this.combineAxes( controlPoints.x, controlPoints.y, controlPoints.z );
	var withGravityPoints = this.combineAxes( controlPoints.xg, controlPoints.yg, controlPoints.zg );

	this.linkPoints( withoutGravityPoints );
	this.linkPoints( withGravityPoints );

	var roots = [];
	if (withoutGravityPoints.length > 0)
	{
		roots[roots.length] = withoutGravityPoints[0];
	}
	if (withGravityPoints.length > 0)
	{
		roots[roots.length] = withGravityPoints[0];
	}
	return roots;
}

/**
 * Link a sorted array of control points.
 *
 * @param controlPoints [array]  an array of control points sorted in ascending timestamp order
 */
Motion.Gesture.Record.Controls.Linker.prototype.linkPoints = function( controlPoints )
{
	var firstControlIndex = 0;
	var lastControlIndex = controlPoints.length - 1;

	for (var i=0; i<controlPoints.length; i++)
	{
		var controlPoint = controlPoints[i];

		controlPoint.previous = (i == firstControlIndex) ? null : controlPoints[i-1];
		controlPoint.next = (i == lastControlIndex) ? null : controlPoints[i+1];
	}
}

/**
 * Build a single array of control points in ascending timestamp order.
 *
 * @param controlPointsX [array]  an array of control points for the x-axis in ascending timestamp order
 * @param controlPointsY [array]  an array of control points for the y-axis in ascending timestamp order
 * @param controlPointsZ [array]  an array of control points for the z-axis in ascending timestamp order
 */
Motion.Gesture.Record.Controls.Linker.prototype.combineAxes = function( controlPointsX, controlPointsY, controlPointsZ )
{
	var combinedPoints = [];

	var xIndex = 0;
	var yIndex = 0;
	var zIndex = 0;

	while (xIndex < controlPointsX.length  ||  yIndex < controlPointsY.length  ||  zIndex < controlPointsZ.length)
	{
		var timestampX = (xIndex < controlPointsX.length) ? controlPointsX[xIndex].timestamp : Infinity;
		var timestampY = (yIndex < controlPointsY.length) ? controlPointsY[yIndex].timestamp : Infinity;
		var timestampZ = (zIndex < controlPointsZ.length) ? controlPointsZ[zIndex].timestamp : Infinity;

		if (   (timestampZ <= timestampX)
		    && (timestampZ <= timestampY)  )
		{
			combinedPoints[combinedPoints.length] = controlPointsZ[zIndex];
			zIndex ++;
		}
		else if (   (timestampY <= timestampX)
			 && (timestampY <= timestampZ)  )
		{
			combinedPoints[combinedPoints.length] = controlPointsY[yIndex];
			yIndex ++;
		}
		else
		{
			combinedPoints[combinedPoints.length] = controlPointsX[xIndex];
			xIndex ++;
		}
	}


	return combinedPoints;
}


/**
 * The merger processes the linked list of control points and where several points are close
 * (on the time axis) it merges those points together.
 * The result is a linked list of controls where each control has many control points.
 *
 * @param mergeTolerance [number]  the maximum number of milli-seconds between control points that should be merged
 */
Motion.Gesture.Record.Controls.Merger = function( mergeTolerance )
{
	this.mergeTolerance = (mergeTolerance) ? mergeTolerance : Motion.Gesture.Record.Controls.Merger.DEFAULT_MERGE_TOLERANCE;
}




/**
 * Merge any touching control points together.
 *
 * @param rootControlPoints [Array(ControlPoint)]  an array of control points that form the root of the gesture signature
 * @return [Array(Control)]  those root control points processed into a set of merged points
 */
Motion.Gesture.Record.Controls.Merger.prototype.merge = function( rootControlPoints )
{
	var mergedArray = [];

	for (var i=0; i<rootControlPoints.length; i++)
	{
		mergedArray[mergedArray.length] = this.mergeLinkedList( rootControlPoints[i] );
	}

	return mergedArray;
}

/**
 * Merge any touching control points together.
 * The control points are processed into controls (multiple control points per control) and
 * returned as a linked list.
 *
 * @param rootControlPoint [ControlPoint]  a root control point
 * @return [Control]  a root control with all subsequent controls attached as a linked list
 */
Motion.Gesture.Record.Controls.Merger.prototype.mergeLinkedList = function( rootControlPoint )
{
	var rootControl = this.createControl( rootControlPoint );

	var controlPoint = rootControlPoint;
	var control = rootControl;
	while (controlPoint.next)
	{
		controlPoint = controlPoint.next;
		if (this.pointShouldBeMerged( control, controlPoint ) === true)
		{
			control = this.mergeControlPoint( control, controlPoint );
		}
		else
		{
			control = this.linkToNewControl( control, controlPoint );
		}
	}

	return rootControl;
}

Motion.Gesture.Record.Controls.Merger.prototype.pointShouldBeMerged = function( control, controlPoint )
{
	if (this.controlAlreadyHasPointInAxis( control, controlPoint.axis ) === true)
	{
		// Don't allow multiple control points in a single axis.
		return false;
	}

	var maxTimestampForMerge = control.timestamp.max + this.mergeTolerance;
	if (controlPoint.timestamp > maxTimestampForMerge)
	{
		// Don't accept control points that are too far from the rest of the points.
		return false;
	}

	return true;
}

Motion.Gesture.Record.Controls.Merger.prototype.controlAlreadyHasPointInAxis = function( control, axis )
{
	for (var i=0; i<control.points.length; i++)
	{
		if (control.points[i].axis === axis)
		{
			return true;
		}
	}

	return false;
}

Motion.Gesture.Record.Controls.Merger.prototype.mergeControlPoint = function( control, controlPoint )
{
	control.points[control.points.length] = controlPoint;

	control.timestamp.max = controlPoint.timestamp;

	return control;
}

Motion.Gesture.Record.Controls.Merger.prototype.linkToNewControl = function( control, controlPoint )
{
	var newControl = this.createControl( controlPoint );

	newControl.previous = control;
	control.next = newControl;

	return newControl;
}

/**
 * Process a raw control point (single point) into a processed control which can have several
 * points merged into one control.
 *
 * @param controlPoint [ControlPoint]  a root control point containing the point information
 * @return [Control]  a processed control, containing an array of control points
 */
Motion.Gesture.Record.Controls.Merger.prototype.createControl = function( controlPoint )
{
	var control = {};
	control.points = [];
	control.points[0] = controlPoint;
	control.timestamp = {};
	control.timestamp.min = controlPoint.timestamp;
	control.timestamp.max = controlPoint.timestamp;

	return control;
}




Motion.Gesture.Record.Controls.Merger.DEFAULT_MERGE_TOLERANCE = 250;


/**
 * The sizer assigns a size (the accepted time window and the allowable range
 * of accelaration value) to each control.
 *
 * Once a control has been passed through one of the 'size' functions the
 * following additional attributes are assigned:
 *  control.window.time.delay           [number]  the number of milliseconds to wait after the previous control point before this control point becomes active
 *  control.window.time.width           [number]  the number of milliseconds this control point is active for
 *  control.points[n].window.value.min  [number]  the minimum acceleration reading to be accepted for this control point
 *  control.points[n].window.value.max  [number]  the maximum acceleration reading to be accepted for this control point
 */
Motion.Gesture.Record.Controls.Sizer = function()
{
	this.maxAccelerationDeviation = 2;
	this.maxTimeDeviation = 250;
}




Motion.Gesture.Record.Controls.Sizer.prototype.setMaxAccelerationDeviation = function( maxAccelerationDeviation )
{
	this.maxAccelerationDeviation = maxAccelerationDeviation;
}

Motion.Gesture.Record.Controls.Sizer.prototype.setMaxTimeDeviation = function( maxTimeDeviation )
{
	this.maxTimeDeviation = maxTimeDeviation;
}

Motion.Gesture.Record.Controls.Sizer.prototype.size = function( rootControls )
{
	for (var i=0; i<rootControls.length; i++)
	{
		this.sizeLinkedList( rootControls[i] );
	}
}

Motion.Gesture.Record.Controls.Sizer.prototype.sizeLinkedList = function( rootControl )
{
	this.sizeControl( rootControl );

	var control = rootControl.next;
	while (control)
	{
		this.sizeControl( control );

		control = control.next;
	}
}

Motion.Gesture.Record.Controls.Sizer.prototype.sizeControl = function( control )
{
	control.window = {};
	control.window.time = this.computeTimeWindow( control );

	for (var i=0; i<control.points.length; i++)
	{
		this.sizePoint( control.points[i] );
	}
}

Motion.Gesture.Record.Controls.Sizer.prototype.sizePoint = function( controlPoint )
{
	controlPoint.window = {};
	controlPoint.window.value = this.computeAccelerationWindow( controlPoint );
}

Motion.Gesture.Record.Controls.Sizer.prototype.computeTimeWindow = function( control )
{
	var timesToAdjacentPoints = this.computeTimesToAdjacentControls( control );
	var timeDeviation = this.computeTimeDeviation( timesToAdjacentPoints );

	var minTime = control.timestamp.min - timeDeviation;
	var maxTime = control.timestamp.max + timeDeviation;

	var timeWindow = {};
	timeWindow.width = maxTime - minTime;
	if (control.previous)
	{
		var previousControlMidPoint = Motion.Gesture.Model.Control.getStartTime( control.previous ) + (control.previous.window.time.width / 2);
		timeWindow.delay = minTime - previousControlMidPoint;
	}
	else
	{
		// The root control has a start time instead of a delay.
		timeWindow.start = minTime;
	}

	return timeWindow;
}

/**
 * Compute the times to the adjacent controls.
 */
Motion.Gesture.Record.Controls.Sizer.prototype.computeTimesToAdjacentControls = function( control )
{
	var timeFromPrevControl;
	if (control.previous)
	{
		timeFromPrevControl = control.timestamp.min - control.previous.timestamp.max;
	}
	else
	{
		timeFromPrevControl = null;
	}

	var timeToNextControl;
	if (control.next)
	{
		timeToNextControl = control.next.timestamp.min - control.timestamp.max;
	}
	else
	{
		timeToNextControl = null;
	}

	return { "fromPrevious": timeFromPrevControl, "toNext": timeToNextControl };
}

/**
 * Compute the time deviation around the control.
 *
 * @param timesToAdjacentControls [object]  the times in msec to the previous and next controls (see computeTimesToAdjacentControls)
 */
Motion.Gesture.Record.Controls.Sizer.prototype.computeTimeDeviation = function( timesToAdjacentControls )
{
	var maxDeviation = Infinity;

	if (timesToAdjacentControls.fromPrevious != null)
	{
		maxDeviation = Math.min( maxDeviation, timesToAdjacentControls.fromPrevious );
	}

	maxDeviation = Math.min( maxDeviation, this.maxTimeDeviation );

	return maxDeviation;
}

/**
 * Compute the min and max acceleration values for the control point.
 */
Motion.Gesture.Record.Controls.Sizer.prototype.computeAccelerationWindow = function( controlPoint )
{
	// This simple implementation is a plain deviation either side of the control point value.

	return { "min": controlPoint.value - this.maxAccelerationDeviation,
		 "max": controlPoint.value + this.maxAccelerationDeviation };
}

if (typeof Motion === "undefined")
	Motion = {};

if (typeof Motion.Gesture === "undefined")
	Motion.Gesture = {};

if (typeof Motion.Gesture.Record === "undefined")
	Motion.Gesture.Record = {};

if (typeof Motion.Gesture.Record.Controls === "undefined")
	Motion.Gesture.Record.Controls = {};




Motion.Gesture.Record.Controls.PointGeneration = {};


/**
 * A strategy class for scoring the locations of a control point.
 * This helps place a new control point at the correct/best timestamp.
 */
Motion.Gesture.Record.Controls.PointGeneration.Scorer = function()
{
}




Motion.Gesture.Record.Controls.PointGeneration.Scorer.prototype.score = function( motionDirection, motionData, motionIndex )
{
	if (motionDirection === Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_INCREASING)
	{
		// Give most weight to an increase in gradient but, where gradient change is about equal, also take into account the min value.
		var gradientChange = this.gradientChange( motionData, motionIndex );
		var value = motionData.values[motionIndex];

		return (gradientChange * 50) - value;
	}
	else if (motionDirection === Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_DECREASING)
	{
		// Give most weight to a decrease in gradient but, where gradient change is about equal, also take into account the max value.
		var gradientChange = this.gradientChange( motionData, motionIndex );
		var value = motionData.values[motionIndex];

		return (gradientChange * -50) + value;
	}
	else if (motionDirection === Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_STEADY)
	{
		// Combination of high magnitude pre gradient and low magnitude post gradient.
		var preGradient = this.preGradient( motionData, motionIndex );
		var postGradient = this.postGradient( motionData, motionIndex );

		return (Math.abs( preGradient ) * 5) - (Math.abs( postGradient ) * -5);
	}

	return 0;
}

Motion.Gesture.Record.Controls.PointGeneration.Scorer.prototype.gradient = function( motionData, firstIndex, secondIndex )
{
	var changeInValue = motionData.values[secondIndex] - motionData.values[firstIndex];
	var changeInTime = motionData.times[secondIndex] - motionData.times[firstIndex];

	return changeInValue / changeInTime;
}

Motion.Gesture.Record.Controls.PointGeneration.Scorer.prototype.preGradient = function( motionData, motionIndex )
{
	if (motionIndex == 0)
	{
		return 0;
	}
	else
	{
		return this.gradient( motionData, motionIndex-1, motionIndex );
	}
}

Motion.Gesture.Record.Controls.PointGeneration.Scorer.prototype.postGradient = function( motionData, motionIndex )
{
	if (motionIndex == motionData.values.length-1)
	{
		return 0;
	}
	else
	{
		return this.gradient( motionData, motionIndex, motionIndex+1 );
	}
}

Motion.Gesture.Record.Controls.PointGeneration.Scorer.prototype.gradientChange = function( motionData, motionIndex )
{
	return this.postGradient( motionData, motionIndex ) - this.preGradient( motionData, motionIndex );
}


/**
 * The generator generates control points from the motion data.
 * This class is abstract. Look for the sub-classes to find an appropriate algorithm.
 *
 * @param stableMotionThreshold {Number}  the expected noise in the motion readings, changes smaller than this value will be ignored
 * @param placementScorer {Motion.Gesture.Record.Controls.PointGeneration.Scorer}  strategy class to help score the potential locations of a control point
 *                                                                                and therefore place the control point at the correct/best location
 */
Motion.Gesture.Record.Controls.PointGeneration.Generator = function()
{
	this.stableMotionThreshold = 0.3;
	this.placementScorer = new Motion.Gesture.Record.Controls.PointGeneration.Scorer();
}




Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.parse = function( motionData )
{
	var gestureTraits = {};

	gestureTraits.x = this.parseAxis( motionData.x, "x" );
	gestureTraits.y = this.parseAxis( motionData.y, "y" );
	gestureTraits.z = this.parseAxis( motionData.z, "z" );
	gestureTraits.xg = this.parseAxis( motionData.xg, "xg" );
	gestureTraits.yg = this.parseAxis( motionData.yg, "yg" );
	gestureTraits.zg = this.parseAxis( motionData.zg, "zg" );

	return gestureTraits;
}

Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.parseMotionDirection = function( previousReading, currentReading )
{
	var change = currentReading - previousReading;
	if (change > this.stableMotionThreshold)
	{
		return Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_INCREASING;
	}
	else if (change < -this.stableMotionThreshold)
	{
		return Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_DECREASING;
	}
	else
	{
		return Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_STEADY;
	}
}

Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.isChangeInMotionAlreadyDetected = function( motionChangeInfo )
{
	return motionChangeInfo != null;
}

Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.isConsistentChange = function( motionChangeInfo, motionDirection )
{
	return motionChangeInfo.direction === motionDirection;
}

Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.createTrait = function( axis, motionData, motionIndex, newMotionDirection )
{
	return { "axis": axis,
		 "timestamp": motionData.times[motionIndex],
		 "value": motionData.values[motionIndex],
		 "direction": newMotionDirection };
}

/**
 * @param motionData {Array(Object)}  the recorded motion data, containing values and times
 * @param centreIndex {Number}  the index of the point in the readings when the search is to take place
 * @param spanIndex {Number}   the number of readings either side of the centreIndex to inspect
 * @param motionDirection {Number}  one of the MOTION DIRECTION constants that defines which direction the motion is moving in
 * @return {Number}  the index of the best value/timestamp to place the control point
 *                   or null if no location can be found
 */
Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.findBestPlacement = function( motionData, centreIndex, spanIndex, motionDirection )
{
	var bestPlacement = null;
	for (var motionIndex=centreIndex-spanIndex; motionIndex<=centreIndex+spanIndex; motionIndex++)
	{
		if (motionIndex>=0  &&  motionIndex<motionData.values.length)
		{
			var placement = { "index": motionIndex,
					  "score": this.placementScorer.score( motionDirection, motionData, motionIndex ) };

			if (bestPlacement == null  ||  placement.score > bestPlacement.score)
			{
				bestPlacement = placement;
			}
		}
	}

	return bestPlacement.index;
}




Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_UNDEFINED = 0;
Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_INCREASING = 1;
Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_STEADY = 2;
Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_DECREASING = 3;


/**
 * A control point generator which looks for a change in motion direction
 * and checks the new direction is consistent before marking a control point
 * at the time the motion direction changed.
 *
 * @param consistentChangeCount [number]  the number of consistent readings before the parser confirms a change of direction
 */
Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChange = function()
{
	this.consistentChangeCount = 3;
}

Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChange.prototype = new Motion.Gesture.Record.Controls.PointGeneration.Generator();




Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChange.prototype.parseAxis = function( motionData, axis )
{
	// We'll run through the motion data and build up a list of traits.
	var traits = [];

	// Start in an unknown/undefined state and the algorithm will pick up the initial direction;
	var currentMotionDirection = Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_UNDEFINED;

	// Start with no change in direction having been detected.
	var motionChangeInfo = null;

	for (var i=1; i<motionData.values.length; i++)
	{
		var motionDirection = this.parseMotionDirection( motionData.values[i-1], motionData.values[i] );

		var motionDirectionHandled = false;
		if (this.isChangeInMotionAlreadyDetected( motionChangeInfo ) === true)
		{
			if (this.isConsistentChange( motionChangeInfo, motionDirection ) === true)
			{
				motionChangeInfo.consistencyCount ++;
				motionDirectionHandled = true;

				if (motionChangeInfo.consistencyCount >= this.consistentChangeCount)
				{
					var trait = this.createTrait( axis, motionData, motionChangeInfo.index, motionChangeInfo.direction );
					if (trait)
					{
						traits[traits.length] = trait;
					}

					currentMotionDirection = motionChangeInfo.direction;
					motionChangeInfo = null;
				}
			}
			else
			{
				// The change in direction has not been consistent.
				// Reset and look again for change.
				motionChangeInfo = null;
			}
		}

		if (motionDirectionHandled === false  &&  motionDirection !== currentMotionDirection)
		{
			// This is the start of a change in motion direction.
			motionChangeInfo = { "direction": motionDirection,
					     "index": i - 1,
					     "consistencyCount": 1 };
		}
		// else
		// The motion is consistent in the same direction. Let it continue.
	}

	return traits;
}


/**
 * A control point generator which looks for a change in motion direction
 * and checks the new direction is consistent before placing a new control point.
 * It then scores values in either direction to find the best place to mark
 * the control point at that change.
 *
 * @param consistentChangeCount [number]  the number of consistent readings before the parser confirms a change of direction
 */
Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChangeWithScoring = function()
{
	this.consistentChangeCount = 2;
}

Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChangeWithScoring.prototype = new Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChange();




Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChangeWithScoring.prototype.createTrait = function( axis, motionData, motionIndex, newMotionDirection )
{
	var bestPlacement = this.findBestPlacement( motionData, motionIndex, this.consistentChangeCount, newMotionDirection );

	if (bestPlacement != null)
	{
		return Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.createTrait.apply( this, [ axis, motionData, bestPlacement, newMotionDirection ] );
	}
	else
	{
		return null;
	}
}


/**
 * A control point generator which looks for a change in motion over several readings - a given distance apart.
 * The exact location of the point is determined by the 'best' value.
 *
 * @param readingDistance [number]  the number of readings to skip before calculating the difference
 */
Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingLongTermChange = function()
{
	this.readingDistance = 3;

	// Because we're combining multiple changes across several readings (readingDistance)
	// this algorithm benefits from a larger threshold for stable motion.
	// I.e. the combined motion from 3 readings is larger than just 1 !
	this.stableMotionThreshold = 1;
}

Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingLongTermChange.prototype = new Motion.Gesture.Record.Controls.PointGeneration.Generator();




Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingLongTermChange.prototype.parseAxis = function( motionData, axis )
{
	// We''ll run through the motion data and build up a list of traits.
	var traits = [];

	// Start in an unknown/undefined state and the algorithm will pick up the initial direction;
	var currentMotionDirection = Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_UNDEFINED;

	// Start with no change in direction having been detected.
	var motionChangeInfo = null;

	for (var i=this.readingDistance; i<motionData.values.length; i++)
	{
		var readFromIndex = i - this.readingDistance;
		var motionDirection = this.parseMotionDirection( motionData.values[readFromIndex], motionData.values[i] );

		if (motionDirection !== currentMotionDirection)
		{
			this.appendNewTrait( traits, axis, motionData, motionChangeInfo );

			// This is the start of a change in motion direction.
			currentMotionDirection = motionDirection;
			motionChangeInfo = { "direction": motionDirection,
					     "index": readFromIndex };
		}
	}

	this.appendNewTrait( traits, axis, motionData, motionChangeInfo );

	return traits;
}

Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingLongTermChange.prototype.appendNewTrait = function( traits, axis, motionData, motionChangeInfo )
{
	if (motionChangeInfo)
	{
		var trait = this.createTrait( axis, motionData, motionChangeInfo.index, motionChangeInfo.direction );
		if (trait)
		{
			traits[traits.length] = trait;
		}
	}
}

Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingLongTermChange.prototype.createTrait = function( axis, motionData, motionIndex, newMotionDirection )
{
	var bestPlacement = this.findBestPlacement( motionData, motionIndex, this.readingDistance, newMotionDirection );

	if (bestPlacement != null)
	{
		return Motion.Gesture.Record.Controls.PointGeneration.Generator.prototype.createTrait.apply( this, [ axis, motionData, bestPlacement, newMotionDirection ] );
	}
	else
	{
		return null;
	}
}

if (typeof Motion === "undefined")
	Motion = {};

if (typeof Motion.Gesture === "undefined")
	Motion.Gesture = {};

if (typeof Motion.Gesture.Record === "undefined")
	Motion.Gesture.Record = {};




Motion.Gesture.Record.Graphs = {};


/**
 * The AxisGraph draws a graph for the motion in a single axis.
 * This base class contains most functionality for the graph. It is responsible for rendering
 * the axis name (title), background and managing the data.
 * However, there are separate sub-classes which take responsibility for rendering the lines.
 * Use them to choose between plain lines (Raphael) and annotated graphs (G Raphael).
 *
 * @param paper [Raphael]  the canvas being used to render the graphs
 * @param title [string]  the name/label for the axis being represented
 * @param x [number]  the x position on the paper of the left edge of the graph
 * @param y [number]  the y position on the paper of the top edge of the graph
 * @param size [object]  an object containing width and height values for the size of the graph
 * @param timeSpan [number]  the number of milli seconds to be shown/buffered on the graph
 */
Motion.Gesture.Record.Graphs.AxisGraph = function( paper, title, x, y, size, timeSpan )
{
	this.paper = paper;
	this.title = title;
	this.x = x;
	this.y = y;
	this.size = size;
	this.timeSpan = timeSpan;
	this.series = [];
	this.data = {};
	this.metadata = {};
	this.graphics = {};
	this.graphics.lines = [];
	this.graphics.backgroundStyle = Motion.Gesture.Record.Graphs.AxisGraph.BACKGROUND_STYLE;
}




Motion.Gesture.Record.Graphs.AxisGraph.prototype.init = function()
{
	this.initBackground();
	this.initTitle();
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.destroy = function()
{
	this.destroyBackground();
	this.destroyTitle();
	this.destroyLines();
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.initLines = function()
{
	for (var i=0; i<this.series.length; i++)
	{
		this.initLine( this.series[i] );
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.destroyLines = function()
{
	for (var i=0; i<this.graphics.lines.length; i++)
	{
		if (this.graphics.lines[i])
		{
			this.graphics.lines[i].remove();
			this.graphics.lines[i] = null;
		}
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.initLine = function( series )
{
	if (!this.graphics.lines[series.id])
	{
		this.graphics.lines[series.id] = this.createLineGraphic( this.data[series.id] );
		// If the data is blank or invalid then the graphic will not be created.
		if (this.graphics.lines[series.id])
		{
			this.graphics.lines[series.id].attr( { stroke: series.style.lineColour, opacity: series.style.lineOpacity } );
		}
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.refreshLine = function( series )
{
	if (this.graphics.lines[series.id])
	{
		this.graphics.lines[series.id].attr( { stroke: series.style.lineColour, opacity: series.style.lineOpacity } );
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.destroyLine = function( series )
{
	if (this.graphics.lines[series.id])
	{
		this.graphics.lines[series.id].remove();
		this.graphics.lines[series.id] = null;
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.initBackground = function()
{
	if (!this.graphics.background)
	{
		this.graphics.background = this.paper.rect( this.x, this.y, this.size.width, this.size.height );
		this.graphics.background.attr( this.graphics.backgroundStyle );
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.destroyBackground = function()
{
	if (this.graphics.background)
	{
		this.graphics.background.remove();
		this.graphics.background = null;
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.initTitle = function()
{
	if (!this.graphics.title)
	{
		this.graphics.title = this.paper.text( this.x, this.y, this.title );
		this.graphics.title.attr( Motion.Gesture.Record.Graphs.AxisGraph.TITLE_STYLE );

		var size = this.graphics.title.getBBox();
		var xPos = this.x + (size.width / 2) + Motion.Gesture.Record.Graphs.AxisGraph.TITLE_HORIZ_MARGIN;
		var yPos = this.y + (size.height / 2);
		this.graphics.title.attr( { x: xPos, y: yPos } );
	}

	if (!this.graphics.titleBackground)
	{
		var size = this.graphics.title.getBBox();
		this.graphics.titleBackground = this.paper.rect( this.x, this.y, size.width + (2 * Motion.Gesture.Record.Graphs.AxisGraph.TITLE_HORIZ_MARGIN), size.height );
		this.graphics.titleBackground.attr( Motion.Gesture.Record.Graphs.AxisGraph.TITLE_BACKGROUND_STYLE );
	}

	this.graphics.title.toFront();
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.destroyTitle = function()
{
	if (this.graphics.title)
	{
		this.graphics.title.remove();
		this.graphics.title = null;
	}
	if (this.graphics.titleBackground)
	{
		this.graphics.titleBackground.remove();
		this.graphics.titleBackground = null;
	}
}

Motion.Gesture.Record.Graphs.AxisGraph.prototype.setData = function( series, data, metadata )
{
	if (typeof( this.series[series] ) == "undefined")
	{
		this.series[series] = series;
	}

	this.destroyLine( series );

	this.data[series.id] = data;
	this.metadata[series.id] = metadata;

	this.initLine( series );
}

/**
 * @param timeSpan [number]  the number of milli seconds to be shown/buffered on the graphs
 */
Motion.Gesture.Record.Graphs.AxisGraph.prototype.setTimeSpan = function( timeSpan )
{
	this.timeSpan = timeSpan;

	this.destroyLines();

	this.initLines();
}

/**
 * Compute the coordinate on the canvas for a point on the graph.
 *
 * @param series [object]  the series model
 * @param xValue [number]  the x-value on the graph
 * @param yValue [number]  the y-value on the graph
 * @return [object]  containing x and y values for the coordinate on the canvas
 */
Motion.Gesture.Record.Graphs.AxisGraph.prototype.getPoint = function( series, xValue, yValue )
{
	var data = this.data[series.id];
	var metadata = this.metadata[series.id];

	var xMax = metadata.record.endTime;
	var xRange = this.timeSpan;
	var xScale = this.size.width / xRange;
	var xPos = this.x + this.size.width - ((xMax - xValue) * xScale);

	var yMax = Motion.Gesture.Record.Graphs.RaphaelAxisGraph.YAXIS_RANGE / 2;
	var yRange = Motion.Gesture.Record.Graphs.RaphaelAxisGraph.YAXIS_RANGE;
	var yScale = this.size.height / yRange;
	var yPos = this.y + ((yMax - yValue) * yScale);

	return { x: xPos, y: yPos };
}




Motion.Gesture.Record.Graphs.AxisGraph.BACKGROUND_STYLE = { fill: "#E0E0E0", stroke: "#C0C0C0" };
Motion.Gesture.Record.Graphs.AxisGraph.TITLE_STYLE = { "font-size": 14, "font-family": "tahoma, helvetica", "font-weight": "bold", opacity: 0.7 };
Motion.Gesture.Record.Graphs.AxisGraph.TITLE_BACKGROUND_STYLE = { fill: "#FFFFFF", opacity: 0.3 };
Motion.Gesture.Record.Graphs.AxisGraph.TITLE_HORIZ_MARGIN = 3;


/**
 * A concrete implementation of Motion.Gesture.Record.Graphs.AxisGraph which uses GRaphael to render the graph.
 */
Motion.Gesture.Record.Graphs.GRaphaelAxisGraph = function( paper, title, x, y, size, timeSpan, data )
{
	Motion.Gesture.Record.Graphs.AxisGraph.apply( this, [paper, title, x, y, size, timeSpan, data] );
}

Motion.Gesture.Record.Graphs.GRaphaelAxisGraph.prototype = new Motion.Gesture.Record.Graphs.AxisGraph();




/**
 * Create the line graph.
 *
 * @param data [object]  a model of the data, containing 'times' and 'values'
 * @return [Raphael.el]  a Raphael Graph element representing the line chart
 */
Motion.Gesture.Record.Graphs.GRaphaelAxisGraph.prototype.createLineGraphic = function( data )
{
	var line = this.paper.linechart( this.x, this.y, this.size.width, this.size.height, data.times, data.values, { nostroke: false, axis: "0 0 0 1", symbol: "circle", smooth: true } ). hoverColumn( function ()
	{
		this.tags = this.paper.set();

		for (var i = 0, ii = this.y.length; i < ii; i++)
		{
			this.tags.push(this.paper.tag(this.x, this.y[i], this.values[i], 160, 10).insertBefore(this).attr([{ fill: "#fff" }, { fill: this.symbols[i].attr("fill") }]));
		}
	}, function ()
	{
		this.tags && this.tags.remove();
	} );
	line.symbols.attr( { r: 3 } );

	return line;
}


/**
 * The graphs object is a single object that shows graphs for all 6 axes of motion.
 * This class displays the title above each set of 3 graphs but then calls down to
 * the AxisGraph class to render each individual graph.
 *
 * @param paper [Raphael]  the Raphael paper element used to render the graphs
 * @param timeSpan [number]  the number of milli seconds to be shown/buffered on the graphs
 */
Motion.Gesture.Record.Graphs.Multiplex = function( paper, timeSpan )
{
	this.paper = paper;
	this.graphType = Motion.Gesture.Record.Graphs.RaphaelAxisGraph;
	this.graphs = {};


	var graphSize = { width: Motion.Gesture.Record.Graphs.Multiplex.DEFAULT_WIDTH, height: Motion.Gesture.Record.Graphs.Multiplex.DEFAULT_HEIGHT };
	var graphPosition = {};

	graphPosition.x = 0;
	graphPosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN;
	this.graphs.x = new this.graphType( this.paper, "X", graphPosition.x, graphPosition.y, graphSize, timeSpan );
	this.graphs.x.init();

	graphPosition.x = 0;
	graphPosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN + graphSize.height + Motion.Gesture.Record.Graphs.Multiplex.GRAPH_VERTICAL_SPACING;
	this.graphs.y = new this.graphType( this.paper, "Y", graphPosition.x, graphPosition.y, graphSize, timeSpan );
	this.graphs.y.init();

	graphPosition.x = 0;
	graphPosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN + (2 * (graphSize.height + Motion.Gesture.Record.Graphs.Multiplex.GRAPH_VERTICAL_SPACING));
	this.graphs.z = new this.graphType( this.paper, "Z", graphPosition.x, graphPosition.y, graphSize, timeSpan );
	this.graphs.z.init();

	graphPosition.x = Motion.Gesture.Record.Graphs.Multiplex.GRAPH_HORIZ_SPACING + graphSize.width;
	graphPosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN;
	this.graphs.xg = new this.graphType( this.paper, "X", graphPosition.x, graphPosition.y, graphSize, timeSpan );
	this.graphs.xg.init();

	graphPosition.x = Motion.Gesture.Record.Graphs.Multiplex.GRAPH_HORIZ_SPACING + graphSize.width;
	graphPosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN + graphSize.height + Motion.Gesture.Record.Graphs.Multiplex.GRAPH_VERTICAL_SPACING;
	this.graphs.yg = new this.graphType( this.paper, "Y", graphPosition.x, graphPosition.y, graphSize, timeSpan );
	this.graphs.yg.init();

	graphPosition.x = Motion.Gesture.Record.Graphs.Multiplex.GRAPH_HORIZ_SPACING + graphSize.width;
	graphPosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN + (2 * (graphSize.height + Motion.Gesture.Record.Graphs.Multiplex.GRAPH_VERTICAL_SPACING));
	this.graphs.zg = new this.graphType( this.paper, "Z", graphPosition.x, graphPosition.y, graphSize, timeSpan );
	this.graphs.zg.init();


	this.signatureAnnotater = new Motion.Gesture.Record.Graphs.SignatureAnnotater( this.paper, this.graphs );


	var titlePosition = {};

	titlePosition.x = graphSize.width / 2;
	titlePosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN / 2;
	this.titleWithoutGravity = this.paper.text( titlePosition.x, titlePosition.y, Motion.Gesture.Record.Graphs.Multiplex.TITLE_WITHOUT_GRAVITY );
	this.titleWithoutGravity.attr( Motion.Gesture.Record.Graphs.Multiplex.TITLE_STYLE );

	titlePosition.x = graphSize.width + Motion.Gesture.Record.Graphs.Multiplex.GRAPH_HORIZ_SPACING + (graphSize.width / 2);
	titlePosition.y = Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN / 2;
	this.titleWithGravity = this.paper.text( titlePosition.x, titlePosition.y, Motion.Gesture.Record.Graphs.Multiplex.TITLE_WITH_GRAVITY );
	this.titleWithGravity.attr( Motion.Gesture.Record.Graphs.Multiplex.TITLE_STYLE );
}




/**
 * Refresh the lines on the graph.
 * NOTE: This does not refresh the signature annotations.
 *
 * @param series [object]  information about the series to be refreshed
 */
Motion.Gesture.Record.Graphs.Multiplex.prototype.refresh = function( series )
{
	this.graphs.x.refreshLine( series );
	this.graphs.y.refreshLine( series );
	this.graphs.z.refreshLine( series );
	this.graphs.xg.refreshLine( series );
	this.graphs.yg.refreshLine( series );
	this.graphs.zg.refreshLine( series );

	// Refreshing the line graphics will push the signature graphics to the back.
	// Ensure they're clickable by bringing them to the front again.
	this.signatureAnnotater.toFront();
}

/**
 * @param series [object]  information about the series to be displayed
 * @param motion [object]  new values for the motion read from the gesture
 */
Motion.Gesture.Record.Graphs.Multiplex.prototype.setData = function( series, motion )
{
	this.graphs.x.setData( series, motion.x, motion.metadata );
	this.graphs.y.setData( series, motion.y, motion.metadata );
	this.graphs.z.setData( series, motion.z, motion.metadata );
	this.graphs.xg.setData( series, motion.xg, motion.metadata );
	this.graphs.yg.setData( series, motion.yg, motion.metadata );
	this.graphs.zg.setData( series, motion.zg, motion.metadata );
}

/**
 * @param gesture [Gesture]  information about the gesture whose signature has just changed
 */
Motion.Gesture.Record.Graphs.Multiplex.prototype.setSignature = function( gesture )
{
	this.signatureAnnotater.setData( gesture );
}

/**
 * @param timeSpan [number]  the number of milli seconds to be shown/buffered on the graphs
 */
Motion.Gesture.Record.Graphs.Multiplex.prototype.setTimeSpan = function( timeSpan )
{
	this.graphs.x.setTimeSpan( timeSpan );
	this.graphs.y.setTimeSpan( timeSpan );
	this.graphs.z.setTimeSpan( timeSpan );
	this.graphs.xg.setTimeSpan( timeSpan );
	this.graphs.yg.setTimeSpan( timeSpan );
	this.graphs.zg.setTimeSpan( timeSpan );

	// The setTimeSpan function will refresh the graphs.
	// We also need to refresh the signature.
	this.signatureAnnotater.refresh();
}

/**
 * @param gesture [object]  object to identify the specific gesture
 */
Motion.Gesture.Record.Graphs.Multiplex.prototype.setGesture = function( gesture )
{
	this.signatureAnnotater.setGesture( gesture );
}




/* Default dimensions of the individual graphs. */
Motion.Gesture.Record.Graphs.Multiplex.DEFAULT_WIDTH = 300;
Motion.Gesture.Record.Graphs.Multiplex.DEFAULT_HEIGHT = 125;

/* Settings for the titles shown above each set of 3 graphs. */
Motion.Gesture.Record.Graphs.Multiplex.TITLE_MARGIN = 20;
Motion.Gesture.Record.Graphs.Multiplex.TITLE_STYLE = { "font-size": 14, "font-weight": "bold", "font-family": "tahoma, helvetica", "text-anchor": "middle" };
Motion.Gesture.Record.Graphs.Multiplex.TITLE_WITH_GRAVITY = "Acceleration With Gravity";
Motion.Gesture.Record.Graphs.Multiplex.TITLE_WITHOUT_GRAVITY = "Acceleration Without Gravity";

/* Settings for spacing between the graphs. */
Motion.Gesture.Record.Graphs.Multiplex.GRAPH_VERTICAL_SPACING = 5;
Motion.Gesture.Record.Graphs.Multiplex.GRAPH_HORIZ_SPACING = 20;


/**
 * A concrete implementation of Motion.Gesture.Record.Graphs.AxisGraph which uses
 * a simple Raphael line/path to render the graph.
 */
Motion.Gesture.Record.Graphs.RaphaelAxisGraph = function( paper, title, x, y, size, timeSpan, data )
{
	Motion.Gesture.Record.Graphs.AxisGraph.apply( this, [paper, title, x, y, size, timeSpan, data] );
}

Motion.Gesture.Record.Graphs.RaphaelAxisGraph.prototype = new Motion.Gesture.Record.Graphs.AxisGraph();




/**
 * Create the line graph.
 *
 * @param data [object]  a model of the data, containing 'times' and 'values'
 * @return [Raphael.el]  a Raphael element representing the line
 *                       or null, if there is no line to show (due to no data)
 */
Motion.Gesture.Record.Graphs.RaphaelAxisGraph.prototype.createLineGraphic = function( data )
{
	if (data.times.length == 0)
		// There is no data, so show no line.
		return null;

	var pathString = "";
	var pointIndex = data.times.length - 1;
	pathString += "M" + this.createPointString( data, pointIndex );
	while (pointIndex >= 0)
	{
		var pointString = this.createPointString( data, pointIndex );
		if (pointString !== null)
		{
			pathString += "L" + pointString;
		}
		pointIndex --;
	}

	var lineGraphic = this.paper.path( pathString );
	lineGraphic.attr( Motion.Gesture.Record.Graphs.RaphaelAxisGraph.LINE_STYLE );

	return lineGraphic;
}

Motion.Gesture.Record.Graphs.RaphaelAxisGraph.prototype.createPointString = function( data, index )
{
	var point = this.createPoint( data, index );

	if (point !== null)
	{
		if (point.x > this.x)
		{
			return "" + point.x + "," + point.y;
		}
		else
		{
			// The point at the given index is outside the graph.
			return null;
		}
	}
	else
	{
		// There is no point with the given index.
		return null;
	}
}

Motion.Gesture.Record.Graphs.RaphaelAxisGraph.prototype.createPoint = function( data, index )
{
	if (index >= data.times.length  ||  index >= data.values.length)
		return null;

	var xMax = data.times[data.times.length - 1];
	var xRange = this.timeSpan;
	var xValue = data.times[index];
	var xScale = this.size.width / xRange;
	var xPos = this.x + this.size.width - ((xMax - xValue) * xScale);

	var yMax = Motion.Gesture.Record.Graphs.RaphaelAxisGraph.YAXIS_RANGE / 2;
	var yRange = Motion.Gesture.Record.Graphs.RaphaelAxisGraph.YAXIS_RANGE;
	var yValue = data.values[index];
	var yScale = this.size.height / yRange;
	var yPos = this.y + ((yMax - yValue) * yScale);

	return { x: xPos, y: yPos };
}




Motion.Gesture.Record.Graphs.RaphaelAxisGraph.LINE_STYLE = {};

/** The range covered by the y-axis.
  * The axis is symmetrical about the 0 value, so the max value is half this amount. */
Motion.Gesture.Record.Graphs.RaphaelAxisGraph.YAXIS_RANGE = 40;


/**
 * Renders an overlay above the graphs to display the signature of the motion
 * in the selected gesture.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater = function( paper, graphs )
{
	this.paper = paper;
	this.graphs = graphs;
	this.selectedGesture = null;
	this.signatures = {};
	this.graphics = [];
}




/**
 * @param gesture [object]  object to identify the specific gesture
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.setGesture = function( gesture )
{
	this.selectedGesture = gesture;

	this.refresh();
}

/**
 * @param gesture [Gesture]  object to identify the specific gesture
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.setData = function( gesture )
{
	this.signatures[gesture.id] = gesture.signature;

	if (gesture == this.selectedGesture)
	{
		this.refresh();
	}
}

Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.toFront = function()
{
	for (var i=0; i<this.graphics.length; i++)
	{
		this.graphics[i].toFront();
	}
}

Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.refresh = function()
{
	this.clear();
	this.draw();
}

Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.destroy = function()
{
	this.clear();
}

/**
 * An internal method to remove the graphics.
 * You should not call this from outside this class.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.clear = function()
{
	for (var i=0; i<this.graphics.length; i++)
	{
		this.graphics[i].remove();
		this.graphics[i] = null;
	}
	this.graphics = [];
}

/**
 * An internal method to draw the graphics.
 * You should not call this from outside this class.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.draw = function()
{
	var signature = this.signatures[this.selectedGesture.id];
	if (signature)
	{
		this.drawSignature( signature );
	}
	// else: there's no signature for this gesture
}

/**
 * An internal method to draw the graphics for the whole signature.
 * You should not call this from outside this class.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.drawSignature = function( signature )
{
	for (var i=0; i<signature.roots.length; i++)
	{
		this.drawControls( signature.roots[i] );
	}
}

/**
 * An internal method to draw the graphics for the a set of controls, starting with the root control.
 * You should not call this from outside this class.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.drawControls = function( rootControl )
{
	this.drawControl( rootControl );

	var control = rootControl;
	while (control.next)
	{
		control = control.next;
		this.drawControl( control );
	}
}

/**
 * An internal method to draw the graphics for a single control.
 * You should not call this from outside this class.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.drawControl = function( control )
{
	var anyLinkDrawn = false;
	for (var i=0; i<control.points.length; i++)
	{
		var linkDrawn = this.drawControlPoint( control, control.points[i] );

		if (i>0)
		{
			this.drawControlPointCombinedLine( control, control.points[i], control.points[i-1] );
		}

		anyLinkDrawn = anyLinkDrawn || linkDrawn;
	}

	// Sometimes a control will not have a point in the same axis as any points in the previous control.
	// In this case no link will have been drawn by the 'drawControlPoint' method.
	if (anyLinkDrawn === false)
	{
		var controlPoint = control.points[0];    // It's safe to assume a control has at least one point.
		var previousControlPoint = this.getPreviousControlPoint( control, controlPoint );
		if (previousControlPoint)
		{
			this.drawControlPointLinkLine( control, controlPoint, previousControlPoint );
		}
		// else: this is a root point, there is no previous control point to link from
	}
}

/**
 * An internal method to draw the graphics for a single control.
 * You should not call this from outside this class.
 *
 * @param control [Control]  the control which owns the control point
 * @param controlPoint [ControlPoint]  the control point to be drawn
 * @return [Boolean]  true if a link to the previous control point has been drawn
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.drawControlPoint = function( control, controlPoint )
{
	var graph = this.graphs[controlPoint.axis];
	if (!graph)
	{
		return;
	}

	var timeWindowStart = Motion.Gesture.Model.Control.getStartTime( control );
	var centre = graph.getPoint( this.selectedGesture, controlPoint.timestamp, controlPoint.value );
	var topLeft = graph.getPoint( this.selectedGesture, timeWindowStart, controlPoint.window.value.max );
	var bottomRight = graph.getPoint( this.selectedGesture, timeWindowStart + control.window.time.width, controlPoint.window.value.min );
	var width = bottomRight.x - topLeft.x;
	var height = bottomRight.y - topLeft.y;

	var window = this.paper.rect( topLeft.x, topLeft.y, width, height );
	this.graphics[this.graphics.length] = window;
	window.attr( Motion.Gesture.Record.Graphs.SignatureAnnotater.WINDOW_STYLE );

	var spot = this.paper.circle( centre.x, centre.y, Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_RADIUS );
	this.graphics[this.graphics.length] = spot;
	spot.attr( Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_STYLE );
	spot.attr( Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_TYPE_STYLE[controlPoint.direction] );

	var isLinkDrawn = false;
	var previousControlPoint = this.getPreviousControlPointInSameAxis( control, controlPoint.axis );
	if (previousControlPoint)
	{
		this.drawControlPointLinkLine( control, controlPoint, previousControlPoint, topLeft );

		isLinkDrawn = true;
	}
	// else: there is no control point in this axis, rely on the caller to draw a link to a control point in a different axis (if it's needed)

	return isLinkDrawn;
}

/**
 * An internal method to draw a line that connects combined/merged control points.
 * You should not call this from outside this class.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.drawControlPointCombinedLine = function( control, controlPoint, prevControlPoint )
{
	var leftValue = Motion.Gesture.Model.Control.getStartTime( control );
	var leftCentre     = this.getPoint(     controlPoint, leftValue,     controlPoint.value );
	var prevLeftCentre = this.getPoint( prevControlPoint, leftValue, prevControlPoint.value );

	var linePathString = "";
	linePathString += "M" + leftCentre.x + "," + leftCentre.y;
	linePathString += "L" + prevLeftCentre.x + "," + prevLeftCentre.y;

	var combineLine = this.paper.path( linePathString );
	this.graphics[this.graphics.length] = combineLine;
	combineLine.attr( Motion.Gesture.Record.Graphs.SignatureAnnotater.COMBINE_LINE_STYLE );
}

/**
 * An internal method to draw a line that connects a control to the previous control point.
 * You should not call this from outside this class.
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.drawControlPointLinkLine = function( control, controlPoint, previousControlPoint, topLeft )
{
	var _topLeft = (topLeft) ? topLeft : this.getPoint( controlPoint, Motion.Gesture.Model.Control.getStartTime( control ), controlPoint.window.value.max );
	var previousCentre = this.getPoint( previousControlPoint, previousControlPoint.timestamp, previousControlPoint.value );

	var linkPathString = "";
	linkPathString += "M" + previousCentre.x + "," + previousCentre.y;
	linkPathString += "L" + previousCentre.x + "," + _topLeft.y;
	linkPathString += "L" + _topLeft.x + "," + _topLeft.y;

	var link = this.paper.path( linkPathString );
	this.graphics[this.graphics.length] = link;
	link.attr( Motion.Gesture.Record.Graphs.SignatureAnnotater.LINK_STYLE );
}

/**
 * @param control [Control]  the control
 * @return [ControlPoint]  a control point of the previous control
 *                         or null if there is no previous control point
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.getPreviousControlPoint = function( control, controlPoint )
{
	var previousPoint = null;
	if (control.previous)
	{
		var preferredAxis = Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK[controlPoint.axis];
		for (var i=0; i<control.previous.points.length; i++)
		{
			if (previousPoint  &&  previousPoint.axis === preferredAxis)
			{
				// Keep the existing previous point, which is in the preferred axis.
			}
			else
			{
				previousPoint = control.previous.points[i];
			}
		}
	}

	return previousPoint;
}

/**
 * @param control [Control]  the control
 * @param axis [String]  the name of the axis
 * @return [ControlPoint]  a control point of the previous control, taken from the given axis
 *                         or null if there is no previous control point in that axis
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.getPreviousControlPointInSameAxis = function( control, axis )
{
	if (control.previous)
	{
		for (var i=0; i<control.previous.points.length; i++)
		{
			if (control.previous.points[i].axis == axis)
			{
				return control.previous.points[i];
			}
		}
	}

	return null;
}

/**
 * An internal method to find the coordinates of a point in the current/selected gesture.
 *
 * @param controlPoint [object]  the control point data model
 * @param time [number]  the x-axis (time) value
 * @param value [number]  the y-axis (acceleration) value
 * @return [object]  an object containing the x and y values for the point on the graphics paper
 *                   or null if the control point axis was invalid or not found
 */
Motion.Gesture.Record.Graphs.SignatureAnnotater.prototype.getPoint = function( controlPoint, time, value )
{
	var graph = this.graphs[controlPoint.axis];
	if (!graph)
	{
		console.log( "MG: No graph found for axis '" + controlPoint.axis + "'." );
		return null;
	}

	return graph.getPoint( this.selectedGesture, time, value );
}




/** Style for the box (window) around the control point. */
Motion.Gesture.Record.Graphs.SignatureAnnotater.WINDOW_STYLE = { "stroke": "#808080", "opacity": 0.7 };

/** Style for the line connecting windows that combine to form a single control. */
Motion.Gesture.Record.Graphs.SignatureAnnotater.COMBINE_LINE_STYLE = { "stroke": "#808080", "stroke-dasharray": "- ", "opacity": 0.3 };

/** Style for the line connecting control points to the next control. */
Motion.Gesture.Record.Graphs.SignatureAnnotater.LINK_STYLE = { "stroke": "#404040", "stroke-width": 2, "opacity": 0.9 };

/** Style for the circle that pin-points the control point. */
Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_RADIUS = 4;
Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_STYLE = { "stroke": "#808080", "opacity": 0.7 };
Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_TYPE_STYLE = [];
Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_TYPE_STYLE[Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_UNDEFINED] = { "stroke": "#808080" };
Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_TYPE_STYLE[Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_INCREASING] = { "stroke": "#606090" };
Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_TYPE_STYLE[Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_STEADY] = { "stroke": "#906060" };
Motion.Gesture.Record.Graphs.SignatureAnnotater.SPOT_TYPE_STYLE[Motion.Gesture.Record.Controls.PointGeneration.Generator.MOTION_DECREASING] = { "stroke": "#609060" };

/** The preferred axis to draw a link line to when the two control points are not in the same axis.
    These are defined so as to minimise the length of lines, thereby reducing clutter on the graphs. */
Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK = {};
Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK["x"] = "y";
Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK["y"] = "x";
Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK["z"] = "y";
Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK["xg"] = "yg";
Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK["yg"] = "xg";
Motion.Gesture.Record.Graphs.SignatureAnnotater.PREFERRED_AXIS_LINK["zg"] = "yg";


/**
 * A class to manage interaction with the user interface.
 * This class assigns click handlers to the UI components and passes the action down to the recorder.
 * This class is flexible in the buttons available on the UI. Pass the ids of the elements in the
 * elementIds object to indicate which buttons are available.
 *
 * You should call the init method after constructing the instance.
 *
 * @param recorder [Motion.Gesture.Record.Recorder]  the recorder object that is managing the recording of gestures
 * @param gestures [Array{Gesture}]  an array of gestures objects, the signatures of which should be empty so that new gestures can be recorded and assigned to those gesture objects
 * @param elementIds [Object]  names of the page elements which should be used to trigger various operations
 *                             the object should contain the following properties:
 *                              > motion [object]  sub-group for elements that control recording motion data
 *                                 > stop [string]  id of the element that triggers the recording of motion to stop
 *                                 > start [string]  id of the element that triggers the recording of motion to start
 *                                 > clear [string]  id of the element that triggers the motion data to be cleared
 *                                 > output [string]  id of the element that triggers the motion data to be presented as text
 *                                 > input [string]  id of the element that triggers the motion data to be read from text
 *                              > gesture [object]  sub-group for elements that control operations of gestures
 *                                 > panel [string]  id of the element that contains the controls
 *                                 > copy [string]  id of the element that triggers the gesture to be copied from the current motion data
 *                                 > clear [string]  id of the element that triggers the gesture to be cleared
 *                                 > opacity [object]  sub-group for elements that control the opacity of the gesture motion line
 *                                    > increase [string]  id of the element that triggers the line's opacity to increase
 *                                    > decrease [string]  id of the element that triggers the line's opacity to decrease
 *                                    > value [string]  id of the element that displays the line's opacity
 */
Motion.Gesture.Record.Interface = function( recorder, gestures, elementIds )
{
	this.recorder = recorder;
	this.gestures = gestures;
	this.elementIds = elementIds;
	this.selectedGesture = null;
}




Motion.Gesture.Record.Interface.prototype.init = function( defaultGesture )
{
	this.associateTrigger( this.elementIds.motion.start, Motion.Gesture.Record.Interface.prototype.handleStartRecording );
	this.associateTrigger( this.elementIds.motion.stop, Motion.Gesture.Record.Interface.prototype.handleStopRecording );
	this.associateTrigger( this.elementIds.motion.clear, Motion.Gesture.Record.Interface.prototype.handleClearMotionData );
	this.associateTrigger( this.elementIds.motion.output, Motion.Gesture.Record.Interface.prototype.handleOutputMotionData );
	this.associateTrigger( this.elementIds.motion.input, Motion.Gesture.Record.Interface.prototype.handleInputMotionData );
	this.associateTrigger( this.elementIds.gesture.test, Motion.Gesture.Record.Interface.prototype.handleTestGesture );
	this.associateTrigger( this.elementIds.gesture.output, Motion.Gesture.Record.Interface.prototype.handleOutputGesture );
	this.associateTrigger( this.elementIds.gesture.input, Motion.Gesture.Record.Interface.prototype.handleInputGesture );
	this.associateTrigger( this.elementIds.gesture.copy, Motion.Gesture.Record.Interface.prototype.handleCopyMotion );
	this.associateTrigger( this.elementIds.gesture.clear, Motion.Gesture.Record.Interface.prototype.handleClearGesture );
	this.associateTrigger( this.elementIds.gesture.opacity.increase, Motion.Gesture.Record.Interface.prototype.handleIncreaseOpacity );
	this.associateTrigger( this.elementIds.gesture.opacity.decrease, Motion.Gesture.Record.Interface.prototype.handleDecreaseOpacity );


	for (var i=0; i<this.gestures.length; i++)
	{
		var gesture = this.gestures[i];

		var tabElement = this.associateTrigger( gesture.ui.tabElementId, Motion.Gesture.Record.Interface.prototype.handleSelectGesture, gesture );
		tabElement.style.backgroundColor = gesture.ui.tabColour;

		this.recorder.setGestureData( gesture );
	}

	this.selectGesture( defaultGesture );


	var thisInterface = this;
	this.recorder.addGestureRecognitionHandler( function( gesture )
	{
		thisInterface.flashGestureRecognised( gesture );
	} );
}

Motion.Gesture.Record.Interface.prototype.associateTrigger = function( elementName, method, params )
{
	var element;

	if (elementName)
	{
		element = document.getElementById( elementName );
		if (element)
		{
			var thisInterface = this;
			element.onclick = function()
			{
				method.apply( thisInterface, [params] );
			};
		}
	}

	return element;
}

Motion.Gesture.Record.Interface.prototype.handleStartRecording = function()
{
	this.recorder.startRecording();
}

Motion.Gesture.Record.Interface.prototype.handleStopRecording = function()
{
	this.recorder.stopRecording();
}

Motion.Gesture.Record.Interface.prototype.handleClearMotionData = function()
{
	this.recorder.clearMotionData();
}

Motion.Gesture.Record.Interface.prototype.handleOutputMotionData = function()
{
	var jsonString = this.recorder.getMotionDataAsJson();

	if (this.elementIds.output  &&  document.getElementById( this.elementIds.output ))
	{
		document.getElementById( this.elementIds.output ).innerHTML = "<h2>LIVE MOTION DATA</h2>" + jsonString;
	}
	else
	{
		alert( jsonString );
	}
}

Motion.Gesture.Record.Interface.prototype.handleInputMotionData = function()
{
	var jsonString = prompt( "Enter the data as JSON:" );
// TODO  Use a styled form to prompt for the JSON.
	if (jsonString != null)
	{
		this.recorder.parseMotionDataFromJson( jsonString );
	}
}

Motion.Gesture.Record.Interface.prototype.handleSelectGesture = function( gesture )
{
	this.selectGesture( gesture );
}

Motion.Gesture.Record.Interface.prototype.selectGesture = function( gesture )
{
	var tabElementName = gesture.ui.tabElementId;
	var panelElementName = this.elementIds.gesture.panel;
	if (tabElementName && panelElementName)
	{
		var tabElement = document.getElementById( tabElementName );
		var panelElement = document.getElementById( panelElementName );
		if (tabElement && panelElement)
		{
			this.selectedGesture = gesture;
			panelElement.style.backgroundColor = gesture.ui.tabColour;
			this.refreshOpacityValue();
		}
	}

	this.recorder.selectGesture( gesture );
}

Motion.Gesture.Record.Interface.prototype.handleTestGesture = function()
{
	if (Motion.Gesture.Model.Gesture.isDefined( this.selectedGesture ))
	{
		this.recorder.testGestureWithMotionData( this.selectedGesture );
	}
}

Motion.Gesture.Record.Interface.prototype.handleOutputGesture = function()
{
	var jsonString = this.recorder.getGestureDefinitionAsJson( this.selectedGesture, Motion.Gesture.Record.Interface.GESTURE_JSON_INDENTATION );

	if (this.elementIds.output  &&  document.getElementById( this.elementIds.output ))
	{
		document.getElementById( this.elementIds.output ).innerHTML = "<h2>GESTURE " + this.selectedGesture.id.toUpperCase() + "</h2><pre>" + jsonString + "</pre>";
	}
	else
	{
		alert( jsonString );
	}
}

Motion.Gesture.Record.Interface.prototype.handleInputGesture = function()
{
	var jsonString = prompt( "Enter the gesture definition as JSON:" );
// TODO  Use a styled form to prompt for the JSON.
	if (jsonString != null)
	{
		this.recorder.parseGestureDefinitionFromJson( this.selectedGesture, jsonString );
	}
}

Motion.Gesture.Record.Interface.prototype.handleCopyMotion = function()
{
	this.recorder.copyMotionToGesture( this.selectedGesture );
}

Motion.Gesture.Record.Interface.prototype.handleClearGesture = function()
{
	this.recorder.clearGesture( this.selectedGesture );
}

Motion.Gesture.Record.Interface.prototype.handleIncreaseOpacity = function()
{
	this.selectedGesture.style.lineOpacity += 0.2;
	if (this.selectedGesture.style.lineOpacity > 1)
	{
		this.selectedGesture.style.lineOpacity = 1;
	}

	this.recorder.refreshGestureGraph( this.selectedGesture );
	this.refreshOpacityValue();
}

Motion.Gesture.Record.Interface.prototype.handleDecreaseOpacity = function()
{
	this.selectedGesture.style.lineOpacity -= 0.2;
	if (this.selectedGesture.style.lineOpacity < 0)
	{
		this.selectedGesture.style.lineOpacity = 0;
	}

	this.recorder.refreshGestureGraph( this.selectedGesture );
	this.refreshOpacityValue();
}

Motion.Gesture.Record.Interface.prototype.refreshOpacityValue = function()
{
	var opacityValueElement = document.getElementById( this.elementIds.gesture.opacity.value );
	if (opacityValueElement)
	{
		opacityValueElement.innerHTML = "" + Math.round( this.selectedGesture.style.lineOpacity * 100 ) + "%";
	}
}

Motion.Gesture.Record.Interface.prototype.flashGestureRecognised = function( gesture )
{
	var tabElement = document.getElementById( gesture.ui.tabElementId );
	if (tabElement)
	{
		tabElement.style.backgroundColor = "#000000";

		var restoreTabColourFunction = function()
		{
			tabElement.style.backgroundColor = gesture.ui.tabColour;
		};
		window.setTimeout( restoreTabColourFunction, 750 );
	}
}




Motion.Gesture.Record.Interface.GESTURE_JSON_INDENTATION = "   ";


/**
 * The motion data manager performs operations on raw motion data.
 * The motion data is held in raw JS objects so that the data can be serialised without the
 * definitions of the operations being serialised too.
 * This manager class holds the operations associated with the raw motion data,
 * such as: clear, append, truncate.
 *
 * @param data [object]  the data model (see the newData method) that is to be managed by this manager
 * @param timeSpan [number]  the number of milli seconds to be held in the data buffer
 * @param autoTruncate [boolean]  whether the data is truncated as new data is appended
 */
Motion.Gesture.Record.MotionDataManager = function( data, timeSpan, autoTruncate )
{
	this.data = data;
	this.timeSpan = timeSpan;
	this.autoTruncate = autoTruncate;
}




Motion.Gesture.Record.MotionDataManager.prototype.setTimeSpan = function( timeSpan )
{
	this.timeSpan = timeSpan;
}

Motion.Gesture.Record.MotionDataManager.prototype.setAutoTruncate = function( autoTruncate )
{
	this.autoTruncate = autoTruncate;
}

/**
 * NB: This manager takes a copy of the data being passed into this method
 *     and continues to manage the existing data model.
 *
 * @param data [object]  data to be copied into the manager's data model
 */
Motion.Gesture.Record.MotionDataManager.prototype.setData = function( data )
{
	this.data.metadata = (data.metadata) ? data.metadata : Motion.Gesture.Record.MotionDataManager.newMetadata();

	this.data.x = data.x;
	this.data.y = data.y;
	this.data.z = data.z;
	this.data.xg = data.xg;
	this.data.yg = data.yg;
	this.data.zg = data.zg;
}

Motion.Gesture.Record.MotionDataManager.prototype.createCopy = function()
{
	return Motion.Gesture.Record.MotionDataManager.createCopy( this.data );
}

Motion.Gesture.Record.MotionDataManager.prototype.clear = function()
{
	this.data.metadata = Motion.Gesture.Record.MotionDataManager.newMetadata();

	this.clearAxis( "x" );
	this.clearAxis( "y" );
	this.clearAxis( "z" );
	this.clearAxis( "xg" );
	this.clearAxis( "yg" );
	this.clearAxis( "zg" );
}

Motion.Gesture.Record.MotionDataManager.prototype.clearAxis = function( axis )
{
	var axisData = this.data[axis];

	axisData.values = [];
	axisData.times = [];
}

Motion.Gesture.Record.MotionDataManager.prototype.append = function( timestamp, values )
{
	if (!this.data.metadata.record.endTime  ||  timestamp > this.data.metadata.record.endTime)
		this.data.metadata.record.endTime = timestamp;

	this.appendToAxis( "x", timestamp, values.x );
	this.appendToAxis( "y", timestamp, values.y );
	this.appendToAxis( "z", timestamp, values.z );
	this.appendToAxis( "xg", timestamp, values.xg );
	this.appendToAxis( "yg", timestamp, values.yg );
	this.appendToAxis( "zg", timestamp, values.zg );
}

Motion.Gesture.Record.MotionDataManager.prototype.appendToAxis = function( axis, timestamp, value )
{
	var axisData = this.data[axis];

	axisData.times[axisData.times.length] = timestamp;
	axisData.values[axisData.values.length] = value;

	if (this.autoTruncate === true)
	{
		this.truncate();
	}
}

Motion.Gesture.Record.MotionDataManager.prototype.truncate = function()
{
	this.truncateAxis( "x" );
	this.truncateAxis( "y" );
	this.truncateAxis( "z" );
	this.truncateAxis( "xg" );
	this.truncateAxis( "yg" );
	this.truncateAxis( "zg" );
}

Motion.Gesture.Record.MotionDataManager.prototype.truncateAxis = function( axis )
{
	var axisData = this.data[axis];

	if (this.timeSpan  &&  axisData.times.length > 1)
	{
		var maxTimeOnGraph = axisData.times[axisData.times.length - 1];
		var minTimeOnGraph = maxTimeOnGraph - this.timeSpan;

		var indexOfFirstDataOnGraph = 0;
		while (axisData.times[indexOfFirstDataOnGraph] < minTimeOnGraph)
		{
			indexOfFirstDataOnGraph ++;
		}

		if (indexOfFirstDataOnGraph > 0)
		{
			axisData.times.splice( 0, indexOfFirstDataOnGraph );
			axisData.values.splice( 0, indexOfFirstDataOnGraph );
		}
		// else
		//   all the data held does appear on the graph
	}
	// else
	//   no time span is defined
	//   or there's no need to truncate because there is no data
	//   or there's no need to truncate because there is only one point (which can't be out-of-date unless the timespan is 0!)
}

Motion.Gesture.Record.MotionDataManager.prototype.dataAsJson = function()
{
	if (!JSON)
	{
		alert( "Your browser does not support JSON processing. Please upgrade to a more recent browser." );
		return null;
	}
	else
	{
		return JSON.stringify( this.data );
	}
}

Motion.Gesture.Record.MotionDataManager.prototype.dataFromJson = function( jsonString )
{
	if (!JSON)
	{
		alert( "Your browser does not support JSON processing. Please upgrade to a more recent browser." );
		return null;
	}
	else
	{
		this.setData( JSON.parse( jsonString ) );
	}
}

/**
 * Iterate through all entries in the recorded motion data. Pass each set of recorded values
 * into the handler.
 *
 * @param {function} dataHandler  The function that will operate on each record data entry.
 * @param {number} startTime  Optional parameter to only process data recorded after a given time.
 */
Motion.Gesture.Record.MotionDataManager.prototype.iterate = function( dataHandler, startTime )
{
	var _startTime = (startTime) ? startTime : 0;

	// Use the x axis to determine how many data points we've recorded,
	// but we assume all axes to have the same number of points.
	var dataLength = this.data.x.times.length;
	for (var i=0; i<dataLength; i++)
	{
		var timestamp = this.data.x.times[i];

		if (timestamp > _startTime)
		{
			var data = { "timestamp": timestamp,
						 "x": this.data.x.values[i],
						 "y": this.data.y.values[i],
						 "z": this.data.z.values[i],
						 "xg": this.data.xg.values[i],
						 "yg": this.data.yg.values[i],
						 "zg": this.data.zg.values[i] };

			dataHandler( data );
		}
	}
}

/*
 * Static methods to initialise new data models which the data manager can manage.
 */

Motion.Gesture.Record.MotionDataManager.newData = function( metadata )
{
	var _metadata = (metadata) ? metadata : Motion.Gesture.Record.MotionDataManager.newMetadata();

	return { "metadata": _metadata,
		 "x": Motion.Gesture.Record.MotionDataManager.newAxisData(),
		 "y": Motion.Gesture.Record.MotionDataManager.newAxisData(),
		 "z": Motion.Gesture.Record.MotionDataManager.newAxisData(),
		 "xg": Motion.Gesture.Record.MotionDataManager.newAxisData(),
		 "yg": Motion.Gesture.Record.MotionDataManager.newAxisData(),
		 "zg": Motion.Gesture.Record.MotionDataManager.newAxisData() };
}

Motion.Gesture.Record.MotionDataManager.newMetadata = function()
{
	return { "record": { "endTime": null } };
}

Motion.Gesture.Record.MotionDataManager.newAxisData = function()
{
	return { "values": [], "times": [] };
}

Motion.Gesture.Record.MotionDataManager.createCopy = function( data )
{
	return { "metadata": data.metadata,
		 "x": Motion.Gesture.Record.MotionDataManager.createAxisCopy( data.x ),
		 "y": Motion.Gesture.Record.MotionDataManager.createAxisCopy( data.y ),
		 "z": Motion.Gesture.Record.MotionDataManager.createAxisCopy( data.z ),
		 "xg": Motion.Gesture.Record.MotionDataManager.createAxisCopy( data.xg ),
		 "yg": Motion.Gesture.Record.MotionDataManager.createAxisCopy( data.yg ),
		 "zg": Motion.Gesture.Record.MotionDataManager.createAxisCopy( data.zg ) };
}

Motion.Gesture.Record.MotionDataManager.createAxisCopy = function( axisData )
{
	return { "values": Motion.Gesture.Record.MotionDataManager.createArrayCopy( axisData.values ),
		 "times": Motion.Gesture.Record.MotionDataManager.createArrayCopy( axisData.times ) };
}

Motion.Gesture.Record.MotionDataManager.createArrayCopy = function( array )
// TODO Move this into the Array class.
{
	var copy = [];

	for (var i=0; i<array.length; i++)
	{
		copy[i] = array[i];
	}

	return copy;
}


/**
 * A motion recorder listens to the motion events from the handheld device and uses graphs to
 * plot the data onto the screen in real time. It also displays the data for recorded gestures.
 *
 * @param paper [Raphael]  the Raphael graphics object used to render the graphs
 * @param optionsArg [object]  an optional object containing fields for the following options:
 *                             recordingPeriod : The time between taking recordings of motion data.
 *                             timeSpan : The number of milli seconds to be shown/buffered on the graphs.
 *                             autoTruncate : Whether the recorded data should be truncated as new data is appended.
 *                             refreshPeriod: The time between refreshing (redrawing) the live data graph.
 *                             recognitionPeriod: The time between sending the motion data through the recognition engine.
 */
Motion.Gesture.Record.Recorder = function( paper, optionsArg )
{
	var options = (optionsArg) ? optionsArg : {};
	var timeSpan = (options.timeSpan) ? options.timeSpan : Motion.Gesture.Record.Recorder.DEFAULT_TIME_SPAN;
	var autoTruncate = (options.autoTruncate) ? options.autoTruncate : Motion.Gesture.Record.Recorder.DEFAULT_TIME_SPAN;

	this.recordingPeriod = (options.recordingPeriod) ? options.recordingPeriod : Motion.Gesture.Record.Recorder.DEFAULT_RECORDING_PERIOD;
	this.refreshPeriod = (options.refreshPeriod) ? options.refreshPeriod : Motion.Gesture.Record.Recorder.DEFAULT_REFRESH_PERIOD;
	this.recognitionPeriod = (options.recognitionPeriod) ? options.recognitionPeriod : Motion.Gesture.Record.Recorder.DEFAULT_RECOGNITION_PERIOD;

	this.motionData = Motion.Gesture.Record.MotionDataManager.newData();
	this.motionDataChanged = false;
	this.motionDataManager = new Motion.Gesture.Record.MotionDataManager( this.motionData, timeSpan, autoTruncate );
	this.graphs = new Motion.Gesture.Record.Graphs.Multiplex( paper, timeSpan );
	this.gestureBuilder = new Motion.Gesture.Record.Controls.Builder( new Motion.Gesture.Record.Controls.PointGeneration.GeneratorUsingConsistentChangeWithScoring(), new Motion.Gesture.Record.Controls.Linker(), new Motion.Gesture.Record.Controls.Merger(), new Motion.Gesture.Record.Controls.Sizer(), new Motion.Gesture.Record.Controls.Cleaner() );
	this.gestureRecogniser = new Motion.Gesture.Recognition.Monitor();
	this.lastRecognitionTimestamp = 0;
	this.gestureRecognitionHandlers = [];

	this.startRecording();
	this.startRefreshing();
}




Motion.Gesture.Record.Recorder.prototype.destroy = function()
{
	this.stopRecording();
	this.stopRefreshing();
}

Motion.Gesture.Record.Recorder.prototype.startRefreshing = function()
{
	if (this.refreshInterval)
		return;

	var thisRecorder = this;
	var refreshFunction = function()
	{
		thisRecorder.refreshLiveDataGraph();
	}

	this.refreshInterval = window.setInterval( refreshFunction, this.refreshPeriod );
}

Motion.Gesture.Record.Recorder.prototype.refreshLiveDataGraph = function()
{
	if (this.motionDataChanged === true)
	{
		this.graphs.setData( Motion.Gesture.Record.Recorder.MOTION_DATA_SERIES, this.motionData );
		this.motionDataChanged = false;
	}
}

Motion.Gesture.Record.Recorder.prototype.stopRefreshing = function()
{
	if (this.refreshInterval)
	{
		window.clearInterval( this.refreshInterval );
		this.refreshInterval = null;
	}
}

Motion.Gesture.Record.Recorder.prototype.startRecording = function()
{
	if (this.recordingInterval)
		this.stopRecording();

	var thisRecorder = this;
	var recordFunction = function()
	{
		thisRecorder.recordMotionData();
	}

	this.recordingInterval = window.setInterval( recordFunction, this.recordingPeriod );

	this.startRecognising();
}

Motion.Gesture.Record.Recorder.prototype.stopRecording = function()
{
	if (this.recordingInterval)
	{
		window.clearInterval( this.recordingInterval );
		this.recordingInterval = null;
	}

	this.stopRecognising();
}

Motion.Gesture.Record.Recorder.prototype.startRecognising = function()
{
	if (this.recognitionInterval)
		return;

	var thisRecorder = this;
	var recogniseFunction = function()
	{
		thisRecorder.recogniseFromLiveData();
	}

	this.recognitionInterval = window.setInterval( recogniseFunction, this.recognitionPeriod );
}

Motion.Gesture.Record.Recorder.prototype.recogniseFromLiveData = function()
{
	var thisRecorder = this;

	// Now feed motion data since the last timestamp through the gestureRecogniser.
	this.motionDataManager.iterate( function( data )
					{
					    thisRecorder.gestureRecogniser.process( data );
					    thisRecorder.lastRecognitionTimestamp = data.timestamp;
					},
					thisRecorder.lastRecognitionTimestamp );
}

Motion.Gesture.Record.Recorder.prototype.stopRecognising = function()
{
	this.gestureRecogniser.stopAllThreads();

	if (this.recognitionInterval)
	{
		window.clearInterval( this.recognitionInterval );
		this.recognitionInterval = null;
	}
}




/* SETTERS & GETTERS */

/**
 * @param timeSpan [number]  the number of milli seconds to be shown/buffered on the graphs
 */
Motion.Gesture.Record.Recorder.prototype.setTimeSpan = function( timeSpan )
{
	this.motionDataManager.setTimeSpan( timeSpan );
	this.graphs.setTimeSpan( timeSpan );
}

/**
 * @param graphSize [object]  new width and height for the graphs
 */
Motion.Gesture.Record.Recorder.prototype.setGraphSize = function( graphSize )
{
	this.graphs.setGraphSize( graphSize );
}

/**
 * @param gestureRecognitionHandler [function]  new function to be called when a gesture is recognised from the live motion data
 */
Motion.Gesture.Record.Recorder.prototype.addGestureRecognitionHandler = function( gestureRecognitionHandler )
{
	this.gestureRecognitionHandlers[this.gestureRecognitionHandlers.length] = gestureRecognitionHandler;
}




/* TOOLS */

Motion.Gesture.Record.Recorder.prototype.getMotionDataAsJson = function()
{
	return this.motionDataManager.dataAsJson();
}

Motion.Gesture.Record.Recorder.prototype.parseMotionDataFromJson = function( jsonString )
{
	this.motionDataManager.dataFromJson( jsonString );

	// Don't set the motionDataChanged flag and wait for the refresh to happen.
	// Display this data immediately,
	this.graphs.setData( Motion.Gesture.Record.Recorder.MOTION_DATA_SERIES, this.motionData );
}

Motion.Gesture.Record.Recorder.prototype.clearMotionData = function()
{
	this.motionDataManager.clear();

	// Don't set the motionDataChanged flag and wait for the refresh to happen.
	// Remove this data immediately,
	this.graphs.setData( Motion.Gesture.Record.Recorder.MOTION_DATA_SERIES, this.motionData );
}

Motion.Gesture.Record.Recorder.prototype.recordMotionData = function()
{
	if (this.deviceMotionData)
	{
		this.motionDataManager.append( new Date().getTime(), this.deviceMotionData );
		this.motionDataChanged = true;
	}
	else
	{
// TODO  Only do this if device does not support the  devicemotion  event.
		// On non-mobile devices load some random data to test the moving graphs.
		var randomData = { x: 10 - (Math.random() * 20),
				   y: 10 - (Math.random() * 20),
				   z: 10 - (Math.random() * 20),
				   xg: 10 - (Math.random() * 20),
				   yg: 10 - (Math.random() * 20),
				   zg: 10 - (Math.random() * 20) };
		this.motionDataManager.append( new Date().getTime(), randomData );
		this.motionDataChanged = true;
	}
}

Motion.Gesture.Record.Recorder.prototype.handleDeviceMotion = function( event )
{
	this.deviceMotionData = { x: event.acceleration.x,
				  y: event.acceleration.y,
				  z: event.acceleration.z,
				  xg: event.accelerationIncludingGravity.x,
				  yg: event.accelerationIncludingGravity.y,
				  zg: event.accelerationIncludingGravity.z };
}

Motion.Gesture.Record.Recorder.prototype.selectGesture = function( gesture )
{
	this.graphs.setGesture( gesture );
}

Motion.Gesture.Record.Recorder.prototype.testGestureWithMotionData = function( gesture )
{
	var gestureRecognition = new Motion.Gesture.Recognition.Monitor();

// TODOA : Show a report on the graph indicating where the match hit the controls.
	gestureRecognition.addGesture( gesture,
				       function()
				       {
					   alert( "Motion data matched gesture " + gesture.id + "." );
				       } );

	// Now feed the recorded motion data through the gestureRecognition object.
	this.motionDataManager.iterate( function( data )
					{
					    gestureRecognition.process( data );
					} );
}

/**
 * @param {Gesture} gesture  The gesture to be converted to JSON.
 * @param {string|undefined} indentation  The padding string used to indent the JSON,
 *                              or leave undefined to serialise the JSON on a single line of text.
 * @returns {string}  The JSON representation of the gesture.
 */
Motion.Gesture.Record.Recorder.prototype.getGestureDefinitionAsJson = function( gesture, indentation )
{
	if (!JSON)
	{
		alert( "Your browser does not support JSON processing. Please upgrade to a more recent browser." );
		return null;
	}
	else
	{
		return Motion.Gesture.Model.Serialisation.signatureToJson( gesture.signature, indentation );
	}
}

Motion.Gesture.Record.Recorder.prototype.parseGestureDefinitionFromJson = function( gesture, jsonString )
{
	if (!JSON)
	{
		alert( "Your browser does not support JSON processing. Please upgrade to a more recent browser." );
		return null;
	}
	else
	{
		gesture.signature = Motion.Gesture.Model.Serialisation.signatureFromJson( jsonString );

		this.setGestureData( gesture );
	}
}

Motion.Gesture.Record.Recorder.prototype.copyMotionToGesture = function( gesture )
{
	var gestureMotionData = this.motionDataManager.createCopy();
	gesture.signature = this.gestureBuilder.build( gestureMotionData );

	this.setGestureData( gesture, gestureMotionData );
}

Motion.Gesture.Record.Recorder.prototype.setGestureData = function( gesture, gestureMotionData )
{
	// There's nothing to show if the gesture does not have a signature.
	if (!gesture.signature)
		return;

	// If the motion data is not defined create blank motion data but with the metadata taken from the gesture signature.
	var _gestureMotionData = (gestureMotionData) ? gestureMotionData : Motion.Gesture.Record.MotionDataManager.newData( gesture.signature.metadata );

	this.graphs.setData( gesture, _gestureMotionData );
	this.graphs.setSignature( gesture );

	var thisRecorder = this;
	this.gestureRecogniser.addGesture( gesture,
					   function()
					   {
						for (var i=0; i<thisRecorder.gestureRecognitionHandlers.length; i++)
						{
							thisRecorder.gestureRecognitionHandlers[i]( gesture );
						}
					   } );
}

Motion.Gesture.Record.Recorder.prototype.clearGesture = function( gesture )
{
	gesture.signature = null;

	this.graphs.setData( gesture, new Motion.Gesture.Record.MotionDataManager.newData() );
	this.graphs.setSignature( gesture );

	this.gestureRecogniser.removeGesture( gesture );
}

Motion.Gesture.Record.Recorder.prototype.refreshGestureGraph = function( gesture )
{
	this.graphs.refresh( gesture );
}




Motion.Gesture.Record.Recorder.DEFAULT_RECORDING_PERIOD = 25;
Motion.Gesture.Record.Recorder.DEFAULT_TIME_SPAN = 5000;
Motion.Gesture.Record.Recorder.DEFAULT_AUTO_TRUNCATE = true;
Motion.Gesture.Record.Recorder.DEFAULT_REFRESH_PERIOD = 100;
Motion.Gesture.Record.Recorder.DEFAULT_RECOGNITION_PERIOD = 100;

Motion.Gesture.Record.Recorder.MOTION_DATA_SERIES = { id: "motion",
					              style: { lineColour: "#000000", lineOpacity: 1 } };
